@@ -1,8 +1,9 @@
 /*
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
- * Copyright (C) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- *  
+ * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
+ * Copyright (c) 2001, 2002, 2003, 2004 The ProFTPD Project team
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -24,29 +25,31 @@
  */
 
 /* Generic configuration and standard header file includes.
- * $Id: conf.h,v 1.11 2001/08/16 20:00:51 flood Exp $
+ * $Id: conf.h,v 1.50 2004/04/29 19:28:24 castaglia Exp $
  */
 
-#ifndef __CONF_H
-#define __CONF_H
+#ifndef PR_CONF_H
+#define PR_CONF_H
+
+#ifndef _GNU_SOURCE
+# define _GNU_SOURCE
+#endif
 
 #include "version.h"
 #include "config.h"
 
 #include "default_paths.h"
 
-/* This is no longer used, autoconf provides us with a top level
- * config.h
- */
-/* #include "platform.h" */
-#include "options.h"
-
 #include <stdlib.h>
 #include <stdio.h>
 #include <stdarg.h>
 #include <pwd.h>
 #include <grp.h>
 
+#ifdef HAVE_STROPTS_H
+# include <stropts.h>
+#endif
+
 #ifdef HAVE_SYS_TYPES_H
 # include <sys/types.h>
 #endif
@@ -55,6 +58,10 @@
 # include <unistd.h>
 #endif
 
+#ifdef HAVE_LIMITS_H
+# include <limits.h>
+#endif
+
 #ifdef HAVE_SYS_SELECT_H
 # include <sys/select.h>
 #endif
@@ -124,6 +131,14 @@ char *strchr(),*strrchr();
 # include <netinet/in.h>
 #endif
 
+#ifdef HAVE_NETINET_IN_SYSTM_H
+# include <netinet/in_systm.h>
+#endif
+
+#ifdef HAVE_NETINET_IP_H
+# include <netinet/ip.h>
+#endif
+
 #ifdef HAVE_NETINET_TCP_H
 # include <netinet/tcp.h>
 #endif
@@ -166,6 +181,9 @@ char *strchr(),*strrchr();
 #endif
 
 #ifdef HAVE_SYS_IOCTL_H
+# ifdef SOLARIS2
+#  define BSD_COMP 1
+# endif
 # include <sys/ioctl.h>
 #endif
 
@@ -181,6 +199,10 @@ char *strchr(),*strrchr();
 # include <fnmatch.h>
 #endif
 
+#ifdef HAVE_UTIME_H
+# include <utime.h>
+#endif
+
 #ifdef HAVE_UTMP_H
 # include <utmp.h>
 #endif
@@ -202,6 +224,12 @@ char *strchr(),*strrchr();
 # endif /* HAVE_SYS_TERMIO_H */
 #endif /* HAVE_SYS_TERMIOS_H */
 
+/* The tunable options header needs to be included after all the system headers,
+ * so that limits are picked up properly.
+ */
+
+#include "options.h"
+
 /* Solaris 2.5 seems to already have a typedef for 'timer_t', so
  * #define timer_t to something else as a workaround.
  */
@@ -210,6 +238,15 @@ char *strchr(),*strrchr();
 # define timer_t p_timer_t
 #endif
 
+/* AIX, when compiled using -D_NO_PROTO, lacks some prototypes without
+ * which ProFTPD may do some funny (and not good) things.  Provide the
+ * prototypes as necessary here.
+ */
+
+#if defined(AIX4) && defined(_NO_PROTO)
+off_t lseek(int, off_t, int);
+#endif
+
 /* See if we have bcopy, if not define them to use the memcpy functions */
 
 #ifndef HAVE_BCOPY
@@ -249,11 +286,6 @@ char *strchr(),*strrchr();
 extern int optopt;
 #endif
 
-/* GNU HURD (and maybe others) do not define MAXPATHLEN */
-#ifndef MAXPATHLEN
-#define MAXPATHLEN      1024
-#endif
-
 /* Necessary for alloca to work */
 #if !defined(__alloca) && !defined(__GNU_LIBRARY__)
 # ifdef __GNUC__
@@ -277,6 +309,12 @@ extern char *alloca();
 
 #endif
 
+#if defined(HAVE_LLU) && SIZEOF_OFF_T == 8
+# define PR_LU	"llu"
+#else
+# define PR_LU	"lu"
+#endif
+
 /********************************************************************
  * This is NOT the user configurable section.  Look in options.h
  * for tunable parameters.
@@ -288,25 +326,75 @@ extern char *alloca();
  * not for the support library.
  */
 
-/* Generic typedefs */
+#ifndef INET_ADDRSTRLEN
+# define INET_ADDRSTRLEN        16
+#endif /* INET_ADDRSTRLEN */
 
-typedef struct in_addr p_in_addr_t;
+#ifndef INET6_ADDRSTRLEN
+# define INET6_ADDRSTRLEN       46
+#endif /* INET6_ADDRSTRLEN */
+
+typedef struct {
+  int na_family;
+
+  /* Note: this assumes that DNS names have a maximum size of
+   * 256 characters.
+   */
+  char na_dnsstr[256];
+  int na_have_dnsstr;
+
+#ifdef USE_IPV6
+  char na_ipstr[INET6_ADDRSTRLEN];
+#else
+  char na_ipstr[INET_ADDRSTRLEN];
+#endif /* USE_IPV6 */
+  int na_have_ipstr;
+
+  /* Note: at some point, this union might/should be replaced with
+   * struct sockaddr_storage.  Why?  The sockaddr_storage struct is
+   * better defined to be aligned on OS/arch boundaries, for more efficient
+   * allocation/access.
+   */
+
+  union {
+    struct sockaddr_in v4;
+#ifdef USE_IPV6
+    struct sockaddr_in6 v6;
+#endif /* USE_IPV6 */
+  } na_addr;
+
+} pr_netaddr_t;
 
 #include "pool.h"
+#include "regexp.h"
 #include "proftpd.h"
 #include "support.h"
 #include "sets.h"
 #include "dirtree.h"
-#include "io.h"
+#include "netio.h"
 #include "modules.h"
+#include "response.h"
 #include "timers.h"
 #include "inet.h"
+#include "child.h"
+#include "netaddr.h"
+#include "netacl.h"
+#include "class.h"
+#include "bindings.h"
 #include "ident.h"
+#include "help.h"
+#include "feat.h"
 #include "ftp.h"
 #include "log.h"
+#include "xferlog.h"
+#include "scoreboard.h"
 #include "data.h"
 #include "libsupp.h"
-#include "fs.h"
+#include "fsio.h"
+#include "mkhome.h"
+#include "ctrls.h"
+#include "event.h"
+#include "pr-syslog.h"
 
 # ifdef HAVE_SETPASSENT
 #  define setpwent()	setpassent(1)
@@ -316,6 +404,15 @@ typedef struct in_addr p_in_addr_t;
 #  define setgrent()	setgroupent(1)
 # endif /* HAVE_SETGROUPENT */
 
+/* Define a buffer size to use for responses, making sure it is big enough
+ * to handle large path names (e.g. for MKD responses).
+ */
+#define PR_RESPONSE_BUFFER_SIZE	 (PR_TUNABLE_BUFFER_SIZE + PR_TUNABLE_PATH_MAX)
+
 #endif /* __PROFTPD_SUPPORT_LIBRARY */
 
-#endif /* __CONF_H */
+#ifdef WITH_DMALLOC
+# include <dmalloc.h>
+#endif /* WITH_DMALLOC */
+
+#endif /* PR_CONF_H */
