@@ -26,7 +26,7 @@
 
 /*
  * Authentication module for ProFTPD
- * $Id: mod_auth.c,v 1.166 2003/10/31 07:22:12 castaglia Exp $
+ * $Id: mod_auth.c,v 1.187 2004/05/19 21:44:26 castaglia Exp $
  */
 
 #include "conf.h"
@@ -42,7 +42,7 @@ extern pid_t mpid;
 
 module auth_module;
 
-static unsigned char auth_create_home = FALSE;
+static unsigned char mkhome = FALSE;
 static int TimeoutLogin = PR_TUNABLE_TIMEOUTLOGIN;
 static int logged_in = 0;
 static int auth_tries = 0;
@@ -56,19 +56,19 @@ static void auth_count_scoreboard(cmd_rec *, char *);
  * particular directory.
  */
 static int lockdown(char *newroot) {
-  log_debug(DEBUG1, "Preparing to chroot() the environment, path = '%s'",
+  pr_log_debug(DEBUG1, "Preparing to chroot() the environment, path = '%s'",
     newroot);
 
   PRIVS_ROOT
   if (pr_fsio_chroot(newroot) == -1) {
     PRIVS_RELINQUISH
-    log_pri(PR_LOG_ERR, "%s chroot(\"%s\"): %s", session.user, newroot,
+    pr_log_pri(PR_LOG_ERR, "%s chroot(\"%s\"): %s", session.user, newroot,
       strerror(errno));
     return -1;
   }
   PRIVS_RELINQUISH
 
-  log_debug(DEBUG1, "Environment successfully chroot()ed.");
+  pr_log_debug(DEBUG1, "Environment successfully chroot()ed.");
   return 0;
 }
 
@@ -92,6 +92,7 @@ static int auth_cmd_chk_cb(cmd_rec *cmd) {
  */
 
 static int auth_login_timeout_cb(CALLBACK_FRAME) {
+  pr_event_generate("core.timeout-login", NULL);
   pr_response_send_async(R_421, "Login Timeout (%d seconds): "
     "closing control connection.", TimeoutLogin);
 
@@ -102,6 +103,7 @@ static int auth_login_timeout_cb(CALLBACK_FRAME) {
 }
 
 static int auth_session_timeout_cb(CALLBACK_FRAME) {
+  pr_event_generate("core.timeout-session", NULL);
   pr_response_send_async(R_421, "Session Timeout (%u seconds): "
     "closing control connection", TimeoutSession);
 
@@ -113,7 +115,7 @@ static int auth_session_timeout_cb(CALLBACK_FRAME) {
 
 static int auth_sess_init(void) {
   config_rec *c = NULL;
-  unsigned char *mkhome = NULL;
+  unsigned char *tmp = NULL;
   int res = 0;
 
   /* Check for a server-specific TimeoutLogin */
@@ -128,30 +130,32 @@ static int auth_sess_init(void) {
   }
 
   PRIVS_ROOT
-  if ((res = pr_open_scoreboard(O_RDWR)) < 0) {
+  res = pr_open_scoreboard(O_RDWR);
+  PRIVS_RELINQUISH
+
+  if (res < 0) {
     switch (res) {
       case PR_SCORE_ERR_BAD_MAGIC:
-        log_debug(DEBUG0, "error opening scoreboard: bad/corrupted file");
+        pr_log_debug(DEBUG0, "error opening scoreboard: bad/corrupted file");
         break;
 
       case PR_SCORE_ERR_OLDER_VERSION:
-        log_debug(DEBUG0, "error opening scoreboard: bad version (too old)");
+        pr_log_debug(DEBUG0, "error opening scoreboard: bad version (too old)");
         break;
 
       case PR_SCORE_ERR_NEWER_VERSION:
-        log_debug(DEBUG0, "error opening scoreboard: bad version (too new)");
+        pr_log_debug(DEBUG0, "error opening scoreboard: bad version (too new)");
         break;
 
       default:
-        log_debug(DEBUG0, "error opening scoreboard: %s", strerror(errno));
+        pr_log_debug(DEBUG0, "error opening scoreboard: %s", strerror(errno));
         break;
     }
   }
-  PRIVS_RELINQUISH
 
   /* Create an entry in the scoreboard for this session. */
   if (pr_scoreboard_add_entry() < 0)
-    log_pri(PR_LOG_NOTICE, "notice: unable to add scoreboard entry: %s",
+    pr_log_pri(PR_LOG_NOTICE, "notice: unable to add scoreboard entry: %s",
       strerror(errno));
 
   pr_scoreboard_update_entry(getpid(),
@@ -161,17 +165,16 @@ static int auth_sess_init(void) {
     PR_SCORE_SERVER_LABEL, main_server->ServerName,
     PR_SCORE_CLIENT_ADDR, session.c->remote_addr,
     PR_SCORE_CLIENT_NAME, session.c->remote_name,
-    PR_SCORE_CLASS, (session.class && session.class->name) ?
-      session.class->name: "",
+    PR_SCORE_CLASS, session.class ? session.class->cls_name : "",
     PR_SCORE_BEGIN_SESSION, time(NULL),
     NULL);
 
   /* Should we create the home for a user, if they don't have one? */
-  if ((mkhome = get_param_ptr(main_server->conf, "CreateHome",
-      FALSE)) != NULL && *mkhome == TRUE)
-    auth_create_home = TRUE;
+  tmp = get_param_ptr(main_server->conf, "CreateHome", FALSE);
+  if (tmp != NULL && *tmp == TRUE)
+    mkhome = TRUE;
   else
-    auth_create_home = FALSE;
+    mkhome = FALSE;
 
   /* Scan the scoreboard now, in order to tally up certain values for
    * substituting in any of the Display* file variables.
@@ -183,6 +186,12 @@ static int auth_sess_init(void) {
 
 static int auth_init(void) {
 
+  /* Add the commands handled by this module to the HELP list. */ 
+  pr_help_add(C_USER, "<sp> username", TRUE);
+  pr_help_add(C_PASS, "<sp> password", TRUE);
+  pr_help_add(C_ACCT, "is not implemented", FALSE);
+  pr_help_add(C_REIN, "is not implemented", FALSE);
+
   /* By default, enable auth checking */
   set_auth_check(auth_cmd_chk_cb);
 
@@ -250,7 +259,7 @@ MODRET auth_post_pass(cmd_rec *cmd) {
 
     if (c->argc == 3) {
       if (!strcmp(c->argv[1], "user")) {
-        if (pr_user_or_expression((char **) &c->argv[2])) {
+        if (pr_expr_eval_user_or((char **) &c->argv[2])) {
 
           if (*((unsigned int *) c->argv[1]) > ctxt_precedence) {
 
@@ -265,7 +274,7 @@ MODRET auth_post_pass(cmd_rec *cmd) {
         }
 
       } else if (!strcmp(c->argv[1], "group")) {
-        if (pr_group_and_expression((char **) &c->argv[2])) {
+        if (pr_expr_eval_group_and((char **) &c->argv[2])) {
 
           if (*((unsigned int *) c->argv[1]) > ctxt_precedence) {
 
@@ -280,8 +289,8 @@ MODRET auth_post_pass(cmd_rec *cmd) {
         }
 
       } else if (!strcmp(c->argv[1], "class")) {
-        if (session.class && session.class->name &&
-            !strcmp(session.class->name, c->argv[2])) {
+        if (session.class &&
+            strcmp(session.class->cls_name, c->argv[2]) == 0) {
 
           if (*((unsigned int *) c->argv[1]) > ctxt_precedence) {
 
@@ -321,7 +330,7 @@ MODRET auth_post_pass(cmd_rec *cmd) {
 
   if (have_user_timeout || have_group_timeout ||
       have_class_timeout || have_all_timeout) {
-    log_debug(DEBUG4, "setting TimeoutSession of %u seconds for current %s",
+    pr_log_debug(DEBUG4, "setting TimeoutSession of %u seconds for current %s",
       TimeoutSession,
       have_user_timeout ? "user" : have_group_timeout ? "group" :
       have_class_timeout ? "class" : "all");
@@ -344,9 +353,9 @@ MODRET auth_post_pass(cmd_rec *cmd) {
      * will need to occur in mod_core's cmd_port() function.
      */
     if (session.c->local_port < 1025)
-      log_debug(DEBUG0, "RootRevoke in effect, disabling active transfers");
+      pr_log_debug(DEBUG0, "RootRevoke in effect, disabling active transfers");
 
-    log_debug(DEBUG0, "RootRevoke in effect, dropped root privs");
+    pr_log_debug(DEBUG0, "RootRevoke in effect, dropped root privs");
   }
 
   return HANDLED(cmd);
@@ -403,33 +412,35 @@ static config_rec *_auth_group(pool *p, char *user, char **group,
   return c;
 }
 
-static config_rec *_auth_anonymous_group(pool *p, char *user)
-{
+static config_rec *_auth_anonymous_group(pool *p, char *user) {
   config_rec *c;
   int ret = 0;
 
-  /* retrieve the session group membership information, so that this check
-   * may work properly
+  /* Retrieve the session group membership information, so that this check
+   * may work properly.
    */
   if (!session.gids && !session.groups &&
       (ret = auth_getgroups(p, user, &session.gids, &session.groups)) < 1)
-    log_debug(DEBUG2, "no supplemental groups found for user '%s'", user);
+    pr_log_debug(DEBUG2, "no supplemental groups found for user '%s'", user);
 
-  c = find_config(main_server->conf,CONF_PARAM,"AnonymousGroup",FALSE);
+  c = find_config(main_server->conf, CONF_PARAM, "AnonymousGroup", FALSE);
 
-  if (c) do {
-    ret = pr_group_and_expression((char**)c->argv);
-  } while (!ret && (c = find_config_next(c,c->next,CONF_PARAM,"AnonymousGroup",FALSE)) != NULL);
+  if (c)
+    do {
+      ret = pr_expr_eval_group_and((char **) c->argv);
+
+    } while (!ret &&
+      (c = find_config_next(c, c->next, CONF_PARAM, "AnonymousGroup",
+        FALSE)) != NULL);
 
   return ret ? c : NULL;
 }
 
-static config_rec *_auth_resolve_user(pool *p,char **user,
-                                      char **ournamep,
-                                      char **anonnamep)
-{
+static config_rec *_auth_resolve_user(pool *p, char **user, char **ournamep,
+    char **anonnamep) {
+
   config_rec *c = NULL, *topc = NULL;
-  char *ourname,*anonname = NULL;
+  char *ourname, *anonname = NULL;
   unsigned char is_alias = FALSE, force_anon = FALSE, *auth_alias_only = NULL;
 
   /* Precendence rules:
@@ -438,7 +449,7 @@ static config_rec *_auth_resolve_user(pool *p,char **user,
    *   3. Normal user login
    */
 
-  ourname = (char*)get_param_ptr(main_server->conf,"UserName",FALSE);
+  ourname = (char *) get_param_ptr(main_server->conf, "UserName", FALSE);
 
   if (ournamep && ourname)
     *ournamep = ourname;
@@ -630,21 +641,19 @@ static char *_get_default_chdir(pool *p, xaset_t *conf) {
       break;
     }
 
-    ret = pr_group_and_expression(((char**)c->argv)+1);
+    ret = pr_expr_eval_group_and(((char **) c->argv)+1);
 
     if (ret) {
       dir = c->argv[0];
       break;
     }
 
-    c = find_config_next(c,c->next,CONF_PARAM,"DefaultChdir",FALSE);
+    c = find_config_next(c, c->next, CONF_PARAM, "DefaultChdir", FALSE);
   }
 
-  /* if the directory is relative, concatenate w/ session.cwd
-   */
-
+  /* If the directory is relative, concatenate w/ session.cwd. */
   if (dir && *dir != '/' && *dir != '~')
-    dir = pdircat(p,session.cwd,dir,NULL);
+    dir = pdircat(p, session.cwd, dir, NULL);
 
   /* Check for any expandable variables. */
   if (dir)
@@ -653,8 +662,7 @@ static char *_get_default_chdir(pool *p, xaset_t *conf) {
   return dir;
 }
 
-/* Determine if the user (non-anon) needs a default root dir
- * other than /
+/* Determine if the user (non-anon) needs a default root dir other than /.
  */
 
 static char *_get_default_root(pool *p) {
@@ -671,7 +679,7 @@ static char *_get_default_root(pool *p) {
       break;
     }
 
-    ret = pr_group_and_expression(((char**)c->argv)+1);
+    ret = pr_expr_eval_group_and(((char **) c->argv)+1);
 
     if (ret) {
       dir = c->argv[0];
@@ -691,21 +699,26 @@ static char *_get_default_root(pool *p) {
 
     else {
       char *realdir;
+      int xerrno = 0;
+
+      /* We need to be the final user here so that if the user has their home
+       * directory with a mode the user proftpd is running (ie the User
+       * directive) as can not traverse down, we can still have the default
+       * root.
+       */
 
-      /*
-      ** We need to be the final user here so that if the user has their home
-      ** directory with a mode the user proftpd is running (ie the User
-      ** directive) as can not traverse down, we can still have the default
-      ** root as ~/public_html/
-      */
       PRIVS_USER
-
-      realdir = dir_realpath(p,dir);
-
+      realdir = dir_realpath(p, dir);
+      if (!realdir)
+        xerrno = errno;
       PRIVS_RELINQUISH
 
       if (realdir)
         dir = realdir;
+
+      else
+        pr_log_pri(PR_LOG_NOTICE, "notice: unable to resolve '%s': %s", dir,
+          strerror(xerrno));
     }
   }
 
@@ -760,18 +773,22 @@ static int _setup_environment(pool *p, char *user, char *pass) {
 
   origuser = user;
   c = _auth_resolve_user(p,&user,&ourname,&anonname);
-  if (c) session.anon_config = c;
+
+  if (c)
+    session.anon_config = c;
 
   if (!user) {
-    log_auth(PR_LOG_NOTICE, "USER %s: user is not a UserAlias from %s [%s] "
+    pr_log_auth(PR_LOG_NOTICE, "USER %s: user is not a UserAlias from %s [%s] "
       "to %s:%i", origuser,session.c->remote_name,
       pr_netaddr_get_ipstr(session.c->remote_addr),
       pr_netaddr_get_ipstr(session.c->local_addr), session.c->local_port);
     goto auth_failure;
   }
 
-  if ((pw = auth_getpwnam(p, user)) == NULL) {
-    log_auth(PR_LOG_NOTICE, "USER %s: no such user found from %s [%s] to %s:%i",
+  pw = auth_getpwnam(p, user);
+  if (pw == NULL) {
+    pr_log_auth(PR_LOG_NOTICE,
+      "USER %s: no such user found from %s [%s] to %s:%i",
       user, session.c->remote_name,
       pr_netaddr_get_ipstr(session.c->remote_addr),
       pr_netaddr_get_ipstr(session.c->local_addr), session.c->local_port);
@@ -792,11 +809,8 @@ static int _setup_environment(pool *p, char *user, char *pass) {
     if ((root_allow = get_param_ptr(c ? c->subset : main_server->conf,
         "RootLogin", FALSE)) == NULL || *root_allow != TRUE) {
       pr_memscrub(pass, strlen(pass));
-      log_auth(PR_LOG_CRIT, "SECURITY VIOLATION: root login attempted.");
+      pr_log_auth(PR_LOG_CRIT, "SECURITY VIOLATION: root login attempted.");
       return 0;
-
-    } else {
-      log_auth(PR_LOG_WARNING, "ROOT FTP login successful.");
     }
   }
 
@@ -814,7 +828,7 @@ static int _setup_environment(pool *p, char *user, char *pass) {
   if (!session.gids && !session.groups &&
       (res = auth_getgroups(p, pw->pw_name, &session.gids,
        &session.groups)) < 1)
-    log_debug(DEBUG2, "no supplemental groups found for user '%s'",
+    pr_log_debug(DEBUG2, "no supplemental groups found for user '%s'",
       pw->pw_name);
 
   /* If c != NULL from this point on, we have an anonymous login */
@@ -863,23 +877,25 @@ static int _setup_environment(pool *p, char *user, char *pass) {
         char errstr[200] = {'\0'};
 
         regerror(re_res, pw_regex, errstr, sizeof(errstr));
-        log_auth(PR_LOG_NOTICE, "ANON %s: AnonRejectPasswords denies login",
+        pr_log_auth(PR_LOG_NOTICE, "ANON %s: AnonRejectPasswords denies login",
           origuser);
+ 
+        pr_event_generate("mod_auth.anon-reject-passwords", session.c);
         goto auth_failure;
       }
     }
 #endif
 
     if (!login_check_limits(c->subset, FALSE, TRUE, &i) || (!aclp && !i) ){
-      log_auth(PR_LOG_NOTICE, "ANON %s (Login failed): Limit access denies "
+      pr_log_auth(PR_LOG_NOTICE, "ANON %s (Login failed): Limit access denies "
         "login.", origuser);
       goto auth_failure;
     }
   }
 
   if (!c && !aclp) {
-    log_auth(PR_LOG_NOTICE, "USER %s (Login failed): Limit access denies login",
-      origuser);
+    pr_log_auth(PR_LOG_NOTICE,
+      "USER %s (Login failed): Limit access denies login", origuser);
     goto auth_failure;
   }
 
@@ -900,7 +916,7 @@ static int _setup_environment(pool *p, char *user, char *pass) {
        */
       if (auth_using_alias && *auth_using_alias == TRUE) {
         user_name = origuser;
-        log_auth(PR_LOG_NOTICE, "ANON AUTH: User %s, Auth Alias %s", user,
+        pr_log_auth(PR_LOG_NOTICE, "ANON AUTH: User %s, Auth Alias %s", user,
           user_name);
       }
     }
@@ -938,22 +954,22 @@ static int _setup_environment(pool *p, char *user, char *pass) {
         break;
 
       case PR_AUTH_NOPWD:
-        log_auth(PR_LOG_NOTICE, "USER %s (Login failed): No such user found.",
-          user);
+        pr_log_auth(PR_LOG_NOTICE,
+          "USER %s (Login failed): No such user found.", user);
         goto auth_failure;
 
       case PR_AUTH_BADPWD:
-        log_auth(PR_LOG_NOTICE, "USER %s (Login failed): Incorrect password.",
-          origuser);
+        pr_log_auth(PR_LOG_NOTICE,
+          "USER %s (Login failed): Incorrect password.", origuser);
         goto auth_failure;
 
       case PR_AUTH_AGEPWD:
-        log_auth(PR_LOG_NOTICE, "USER %s (Login failed): Password expired.",
+        pr_log_auth(PR_LOG_NOTICE, "USER %s (Login failed): Password expired.",
           user);
         goto auth_failure;
 
       case PR_AUTH_DISABLEDPWD:
-        log_auth(PR_LOG_NOTICE, "USER %s (Login failed): Account disabled.",
+        pr_log_auth(PR_LOG_NOTICE, "USER %s (Login failed): Account disabled.",
           user);
         goto auth_failure;
 
@@ -965,6 +981,9 @@ static int _setup_environment(pool *p, char *user, char *pass) {
     if (auth_code < 0)
       goto auth_failure;
 
+    if (pw->pw_uid == PR_ROOT_UID)
+      pr_log_auth(PR_LOG_WARNING, "ROOT FTP login successful.");
+
   } else if (c && (!anon_require_passwd || *anon_require_passwd == FALSE)) {
     session.hide_password = FALSE;
   }
@@ -972,13 +991,13 @@ static int _setup_environment(pool *p, char *user, char *pass) {
   auth_setgrent(p);
 
   if (!auth_check_shell((c ? c->subset : main_server->conf), pw->pw_shell)) {
-    log_auth(PR_LOG_NOTICE, "USER %s (Login failed): Invalid shell: '%s'", user,
-      pw->pw_shell);
+    pr_log_auth(PR_LOG_NOTICE, "USER %s (Login failed): Invalid shell: '%s'",
+      user, pw->pw_shell);
     goto auth_failure;
   }
 
   if (!auth_check_ftpusers((c ? c->subset : main_server->conf), pw->pw_name)) {
-    log_auth(PR_LOG_NOTICE, "USER %s (Login failed): User in " FTPUSERS_PATH,
+    pr_log_auth(PR_LOG_NOTICE, "USER %s (Login failed): User in " FTPUSERS_PATH,
       user);
     goto auth_failure;
   }
@@ -1001,7 +1020,7 @@ static int _setup_environment(pool *p, char *user, char *pass) {
 
     PRIVS_ROOT
     if ((res = set_groups(p, pw->pw_gid, session.gids)) < 0)
-      log_pri(PR_LOG_ERR, "error: unable to set groups: %s",
+      pr_log_pri(PR_LOG_ERR, "error: unable to set groups: %s",
         strerror(errno));
 
 #ifndef PR_DEVEL_COREDUMP
@@ -1033,7 +1052,7 @@ static int _setup_environment(pool *p, char *user, char *pass) {
     /* return all privileges back to that of the daemon, for now */
     PRIVS_ROOT
     if ((res = set_groups(p, daemon_gid, daemon_gids)) < 0)
-      log_pri(PR_LOG_ERR, "error: unable to set groups: %s",
+      pr_log_pri(PR_LOG_ERR, "error: unable to set groups: %s",
         strerror(errno));
 
 #ifndef PR_DEVEL_COREDUMP
@@ -1057,7 +1076,7 @@ static int _setup_environment(pool *p, char *user, char *pass) {
 
       PRIVS_RELINQUISH
 
-      log_pri(PR_LOG_ERR, "changing from %s back to daemon uid/gid: %s",
+      pr_log_pri(PR_LOG_ERR, "changing from %s back to daemon uid/gid: %s",
             session.user, strerror(errno));
 
       end_login(1);
@@ -1070,8 +1089,8 @@ static int _setup_environment(pool *p, char *user, char *pass) {
       session.anon_user = pstrdup(session.pool, pass);
 
     if (!session.chroot_path) {
-      log_pri(PR_LOG_ERR, "%s: Directory %s is not accessible.", session.user,
-        c->name);
+      pr_log_pri(PR_LOG_ERR, "%s: Directory %s is not accessible.",
+        session.user, c->name);
       pr_response_add_err(R_530, "Unable to set anonymous privileges.");
       goto auth_failure;
     }
@@ -1112,7 +1131,7 @@ static int _setup_environment(pool *p, char *user, char *pass) {
 
   /* Create the home directory, if need be. */
 
-  if (!c && auth_create_home) {
+  if (!c && mkhome) {
     if (create_home(p, session.cwd, origuser, pw->pw_uid, pw->pw_gid) < 0) {
 
       /* NOTE: should this cause the login to fail? */
@@ -1132,7 +1151,7 @@ static int _setup_environment(pool *p, char *user, char *pass) {
 
   if (!login_check_limits((c ? c->subset : main_server->conf), FALSE, TRUE,
       &i)) {
-    log_auth(PR_LOG_NOTICE, "%s %s: Limit access denies login.",
+    pr_log_auth(PR_LOG_NOTICE, "%s %s: Limit access denies login.",
       (c != NULL) ? "ANON" : C_USER, origuser);
     goto auth_failure;
   }
@@ -1147,7 +1166,7 @@ static int _setup_environment(pool *p, char *user, char *pass) {
   if (c && c->subset)
     resolve_anonymous_dirs(c->subset);
 
-  log_auth(PR_LOG_NOTICE, "%s %s: Login successful.",
+  pr_log_auth(PR_LOG_NOTICE, "%s %s: Login successful.",
     (c != NULL) ? "ANON" : C_USER, origuser);
 
   /* Write the login to wtmp.  This must be done here because we won't
@@ -1192,12 +1211,12 @@ static int _setup_environment(pool *p, char *user, char *pass) {
   }
 
   if (strcasecmp(xferlog, "NONE") == 0)
-    log_open_xfer(NULL);
+    xferlog_open(NULL);
   else
-    log_open_xfer(xferlog);
+    xferlog_open(xferlog);
 
   if ((res = set_groups(p, pw->pw_gid, session.gids)) < 0)
-    log_pri(PR_LOG_ERR, "error: unable to set groups: %s",
+    pr_log_pri(PR_LOG_ERR, "error: unable to set groups: %s",
       strerror(errno));
 
   PRIVS_RELINQUISH
@@ -1208,7 +1227,7 @@ static int _setup_environment(pool *p, char *user, char *pass) {
     ensure_open_passwd(p);
 
     if (lockdown(defroot) == -1) {
-      log_pri(PR_LOG_ERR, "error: unable to set default root directory");
+      pr_log_pri(PR_LOG_ERR, "error: unable to set default root directory");
       pr_response_send(R_530, "Login incorrect.");
       end_login(1);
     }
@@ -1231,7 +1250,7 @@ static int _setup_environment(pool *p, char *user, char *pass) {
     ensure_open_passwd(p);
 
   if (c && lockdown(session.chroot_path) == -1) {
-    log_pri(PR_LOG_ERR, "error: unable to set anonymous privileges");
+    pr_log_pri(PR_LOG_ERR, "error: unable to set anonymous privileges");
     pr_response_send(R_530, "Login incorrect.");
     end_login(1);
   }
@@ -1264,7 +1283,7 @@ static int _setup_environment(pool *p, char *user, char *pass) {
      geteuid() != pw->pw_uid) {
 
     PRIVS_RELINQUISH
-    log_pri(PR_LOG_ERR, "error: %s setregid() or setreuid(): %s",
+    pr_log_pri(PR_LOG_ERR, "error: %s setregid() or setreuid(): %s",
       session.user, strerror(errno));
     pr_response_send(R_530, "Login incorrect.");
 
@@ -1274,7 +1293,7 @@ static int _setup_environment(pool *p, char *user, char *pass) {
 
   /* If the home directory is NULL or "", reject the login. */
   if (pw->pw_dir == NULL || !strcmp(pw->pw_dir, "")) {
-    log_pri(PR_LOG_ERR, "error: user %s home directory is NULL or \"\"",
+    pr_log_pri(PR_LOG_ERR, "error: user %s home directory is NULL or \"\"",
       session.user);
     pr_response_send(R_530, "Login incorrect.");
     end_login(1);
@@ -1305,12 +1324,12 @@ static int _setup_environment(pool *p, char *user, char *pass) {
 
     if (session.chroot_path != NULL || defroot) {
 
-      log_debug(DEBUG2, "unable to chdir to %s (%s), defaulting to chroot "
+      pr_log_debug(DEBUG2, "unable to chdir to %s (%s), defaulting to chroot "
         "directory %s", session.cwd, strerror(errno),
         (session.chroot_path ? session.chroot_path : defroot));
 
       if (pr_fsio_chdir_canon("/", !showsymlinks) == -1) {
-        log_pri(PR_LOG_ERR, "%s chdir(\"/\"): %s", session.user,
+        pr_log_pri(PR_LOG_ERR, "%s chdir(\"/\"): %s", session.user,
           strerror(errno));
         pr_response_send(R_530,"Login incorrect.");
         end_login(1);
@@ -1321,12 +1340,12 @@ static int _setup_environment(pool *p, char *user, char *pass) {
       /* If we've got defchdir, failure is ok as well, simply switch to
        * user's homedir.
        */
-      log_debug(DEBUG2, "unable to chdir to %s (%s), defaulting to home "
+      pr_log_debug(DEBUG2, "unable to chdir to %s (%s), defaulting to home "
         "directory %s", session.cwd, strerror(errno), pw->pw_dir);
 
       if (pr_fsio_chdir_canon(pw->pw_dir, !showsymlinks) == -1) {
-        log_pri(PR_LOG_ERR, "%s chdir(\"%s\"): %s", session.user, session.cwd,
-          strerror(errno));
+        pr_log_pri(PR_LOG_ERR, "%s chdir(\"%s\"): %s", session.user,
+          session.cwd, strerror(errno));
         pr_response_send(R_530, "Login incorrect.");
         end_login(1);
       }
@@ -1336,7 +1355,7 @@ static int _setup_environment(pool *p, char *user, char *pass) {
       /* Unable to switch to user's real home directory, which is not
        * allowed.
        */
-      log_pri(PR_LOG_ERR, "%s chdir(\"%s\"): %s", session.user, session.cwd,
+      pr_log_pri(PR_LOG_ERR, "%s chdir(\"%s\"): %s", session.user, session.cwd,
         strerror(errno));
       pr_response_send(R_530, "Login incorrect.");
       end_login(1);
@@ -1444,24 +1463,31 @@ static void auth_scan_scoreboard(void) {
   char config_class_users[128] = {'\0'};
   char curr_server_addr[80] = {'\0'};
   xaset_t *conf = NULL;
-  int have_class = (session.class && session.class->name) ? TRUE : FALSE;
 
   snprintf(curr_server_addr, sizeof(curr_server_addr), "%s:%d",
     pr_netaddr_get_ipstr(main_server->addr), main_server->ServerPort);
   curr_server_addr[sizeof(curr_server_addr)-1] = '\0';
 
   /* Determine how many users are currently connected */
-  pr_rewind_scoreboard();
+  if (pr_rewind_scoreboard() < 0)
+    pr_log_pri(PR_LOG_NOTICE, "error rewinding scoreboard: %s",
+      strerror(errno));
+
   while ((score = pr_scoreboard_read_entry()) != NULL) {
 
     /* Make sure it matches our current server */
     if (strcmp(score->sce_server_addr, curr_server_addr) == 0) {
       cur++;
 
+      /* Only count up authenticated clients, as per the documentation. */
+      if (strcmp(score->sce_user, "(none)") == 0)
+        continue;
+
       /* Note: the class member of the scoreboard entry will never be
        * NULL.  At most, it may be the empty string.
        */
-      if (have_class && strcasecmp(score->sce_class, session.class->name) == 0)
+      if (session.class &&
+          strcasecmp(score->sce_class, session.class->cls_name) == 0)
         ccur++;
     }
   }
@@ -1476,12 +1502,12 @@ static void auth_scan_scoreboard(void) {
   c->argv[0] = pcalloc(c->pool, sizeof(unsigned int));
   *((unsigned int *) c->argv[0]) = cur;
 
-  if (have_class) {
+  if (session.class) {
     remove_config(CURRENT_CONF, "CURRENT-CLASS", FALSE);
-    add_config_param_set(&conf, "CURRENT-CLASS", 1, session.class->name);
+    add_config_param_set(&conf, "CURRENT-CLASS", 1, session.class->cls_name);
 
     snprintf(config_class_users, sizeof(config_class_users),
-      "CURRENT-CLIENTS-CLASS-%s", session.class->name);
+      "CURRENT-CLIENTS-CLASS-%s", session.class->cls_name);
     remove_config(CURRENT_CONF, config_class_users, FALSE);
     c = add_config_param_set(&conf, config_class_users, 1, NULL);
     c->argv[0] = pcalloc(c->pool, sizeof(unsigned int));
@@ -1494,21 +1520,6 @@ static void auth_count_scoreboard(cmd_rec *cmd, char *user) {
   long cur = 0, hcur = 0, ccur = 0, hostsperuser = 1, usersessions = 0;
   config_rec *c = NULL, *anon_config = NULL, *maxc = NULL;
   char *origuser, config_class_users[128] = {'\0'};
-  unsigned char classes_enabled = FALSE,
-    *class_engine = get_param_ptr(main_server->conf, "Classes", FALSE);
-
-  if (class_engine && *class_engine == TRUE)
-    classes_enabled = TRUE;
-
-  /* Note: there is an assumption here that if Classes have been enabled,
-   * there will be a corresponding Class defined.  This can cause a
-   * SIGSEGV if not caught.
-   *
-   * The catch is this: if Classes are enabled, but find_class() returns
-   *  NULL, act as if Classes are disabled.
-   */
-  if (classes_enabled && session.class == NULL)
-      classes_enabled = FALSE;
 
   /* Determine how many users are currently connected. */
   origuser = user;
@@ -1522,7 +1533,10 @@ static void auth_count_scoreboard(cmd_rec *cmd, char *user) {
       pr_netaddr_get_ipstr(main_server->addr), main_server->ServerPort);
     curr_server_addr[sizeof(curr_server_addr)-1] = '\0';
 
-    pr_rewind_scoreboard();
+    if (pr_rewind_scoreboard() < 0)
+      pr_log_pri(PR_LOG_NOTICE, "error rewinding scoreboard: %s",
+        strerror(errno));
+
     while ((score = pr_scoreboard_read_entry()) != NULL) {
       unsigned char same_host = FALSE;
 
@@ -1539,6 +1553,10 @@ static void auth_count_scoreboard(cmd_rec *cmd, char *user) {
           if (c && c->config_type == CONF_ANON && cur == 0)
               cur = 1;
 
+          /* Only count authenticated clients, as per the documentation. */
+          if (strcmp(score->sce_user, "(none)") == 0)
+            continue;
+
           cur++;
 
           /* Count up sessions on a per-host basis. */
@@ -1567,8 +1585,8 @@ static void auth_count_scoreboard(cmd_rec *cmd, char *user) {
           }
         }
 
-        if (classes_enabled &&
-            !strcasecmp(score->sce_class, session.class->name))
+        if (session.class &&
+            strcasecmp(score->sce_class, session.class->cls_name) == 0)
           ccur++;
       }
     }
@@ -1581,43 +1599,17 @@ static void auth_count_scoreboard(cmd_rec *cmd, char *user) {
   c->argv[0] = pcalloc(c->pool, sizeof(int));
   *((int *) c->argv[0]) = cur;
 
-  if (classes_enabled) {
+  if (session.class) {
     remove_config(cmd->server->conf, "CURRENT-CLASS", FALSE);
     add_config_param_set(&cmd->server->conf, "CURRENT-CLASS", 1,
-                         session.class->name);
+      session.class->cls_name);
 
     snprintf(config_class_users, sizeof(config_class_users), "%s-%s",
-             "CURRENT-CLIENTS-CLASS", session.class->name);
+      "CURRENT-CLIENTS-CLASS", session.class->cls_name);
     remove_config(cmd->server->conf, config_class_users, FALSE);
     c = add_config_param_set(&cmd->server->conf, config_class_users, 1, NULL);
     c->argv[0] = pcalloc(c->pool, sizeof(int));
     *((int *) c->argv[0]) = ccur;
-
-    /* Too many users in this class? */
-    if (ccur >= session.class->max_connections) {
-      char *display = NULL;
-
-      if (session.sf_flags & SF_ANON)
-        display = (char*) get_param_ptr(session.anon_config->subset,
-                                        "DisplayGoAway",FALSE);
-
-      if (!display)
-        display = (char*) get_param_ptr(cmd->server->conf,
-                                        "DisplayGoAway",FALSE);
-
-      if (display)
-        core_display_file(R_530, display, NULL);
-      else
-        pr_response_send(R_530, "Too many users in your class, "
-          "please try again later.");
-
-      remove_config(cmd->server->conf, C_PASS, FALSE);
-
-      log_auth(PR_LOG_NOTICE, "Connection refused (max clients for class %s).",
-        session.class->name);
-
-      end_login(0);
-    }
   }
 
   /* Try to determine what MaxClients/MaxHosts limits apply to this session
@@ -1625,6 +1617,41 @@ static void auth_count_scoreboard(cmd_rec *cmd, char *user) {
    * be exceeded.
    */
 
+  maxc = find_config(cmd->server->conf, CONF_PARAM, "MaxClientsPerClass",
+    FALSE);
+  while (session.class && maxc) {
+    char *maxstr = "Sorry, the maximum number of clients (%m) from your class "
+      "are already connected.";
+    unsigned int *max = maxc->argv[1];
+
+    if (strcmp(maxc->argv[0], session.class->cls_name) != 0) {
+      maxc = find_config_next(maxc, maxc->next, CONF_PARAM,
+        "MaxClientsPerClass", FALSE);
+      continue;
+    }
+
+    if (maxc->argc > 2)
+      maxstr = maxc->argv[2];
+
+    if (*max &&
+        ccur > *max) {
+      char maxn[20] = {'\0'};
+
+      pr_event_generate("mod_auth.max-clients-per-class",
+        session.class ? session.class->cls_name : NULL);
+
+      snprintf(maxn, sizeof(maxn), "%u", *max);
+      pr_response_send(R_530, "%s", sreplace(cmd->tmp_pool, maxstr, "%m", maxn,
+        NULL));
+      pr_log_auth(PR_LOG_NOTICE,
+        "Connection refused (max clients %u per class %s).", *max,
+        session.class->cls_name);
+      end_login(0);
+    }
+
+    break;
+  }
+
   if ((maxc = find_config((anon_config ? anon_config->subset :
       cmd->server->conf), CONF_PARAM, "MaxClientsPerHost", FALSE)) != NULL) {
     char *maxstr = "Sorry, the maximum number of clients (%m) from your host "
@@ -1637,11 +1664,13 @@ static void auth_count_scoreboard(cmd_rec *cmd, char *user) {
     if (*max && hcur > *max) {
       char maxn[20] = {'\0'};
 
+      pr_event_generate("mod_auth.max-clients-per-host", session.c);
+
       snprintf(maxn, sizeof(maxn), "%u", *max);
       pr_response_send(R_530, "%s", sreplace(cmd->tmp_pool, maxstr, "%m", maxn,
         NULL));
-      log_auth(PR_LOG_NOTICE, "Connection refused (max clients per host %u).",
-        *max);
+      pr_log_auth(PR_LOG_NOTICE,
+        "Connection refused (max clients per host %u).", *max);
       end_login(0);
     }
   }
@@ -1659,11 +1688,13 @@ static void auth_count_scoreboard(cmd_rec *cmd, char *user) {
     if (*max && usersessions > *max) {
       char maxn[20] = {'\0'};
 
+      pr_event_generate("mod_auth.max-clients-per-user", user);
+
       snprintf(maxn, sizeof(maxn), "%u", *max);
       pr_response_send(R_530, "%s", sreplace(cmd->tmp_pool, maxstr, "%m", maxn,
         NULL));
-      log_auth(PR_LOG_NOTICE, "Connection refused (max clients per user %u).",
-        *max);
+      pr_log_auth(PR_LOG_NOTICE,
+        "Connection refused (max clients per user %u).", *max);
       end_login(0);
     }
   }
@@ -1680,10 +1711,12 @@ static void auth_count_scoreboard(cmd_rec *cmd, char *user) {
     if (*max && cur > *max) {
       char maxn[20] = {'\0'};
 
+      pr_event_generate("mod_auth.max-clients", NULL);
+
       snprintf(maxn, sizeof(maxn), "%u", *max);
       pr_response_send(R_530, "%s", sreplace(cmd->tmp_pool, maxstr, "%m", maxn,
         NULL));
-      log_auth(PR_LOG_NOTICE, "Connection refused (max clients %u).", *max);
+      pr_log_auth(PR_LOG_NOTICE, "Connection refused (max clients %u).", *max);
       end_login(0);
     }
   }
@@ -1700,10 +1733,12 @@ static void auth_count_scoreboard(cmd_rec *cmd, char *user) {
     if (*max && hostsperuser > *max) {
       char maxn[20] = {'\0'};
 
+      pr_event_generate("mod_auth.max-hosts-per-user", user);
+
       snprintf(maxn, sizeof(maxn), "%u", *max);
       pr_response_send(R_530, "%s", sreplace(cmd->tmp_pool, maxstr, "%m", maxn,
         NULL));
-      log_auth(PR_LOG_NOTICE, "Connection refused (max hosts per host %u).",
+      pr_log_auth(PR_LOG_NOTICE, "Connection refused (max hosts per host %u).",
         *max);
       end_login(0);
     }
@@ -1712,6 +1747,9 @@ static void auth_count_scoreboard(cmd_rec *cmd, char *user) {
 
 MODRET auth_pre_user(cmd_rec *cmd) {
 
+  if (logged_in)
+    return DECLINED(cmd);
+
   /* Close the passwd and group databases, because libc won't let us see new
    * entries to these files without this (only in PersistentPasswd mode).
    */
@@ -1720,7 +1758,7 @@ MODRET auth_pre_user(cmd_rec *cmd) {
 
   /* Check for a user name that exceeds PR_TUNABLE_LOGIN_MAX. */
   if (strlen(cmd->arg) > PR_TUNABLE_LOGIN_MAX) {
-    log_pri(PR_LOG_NOTICE, "USER %s (Login failed): "
+    pr_log_pri(PR_LOG_NOTICE, "USER %s (Login failed): "
       "maximum login length exceeded", cmd->arg);
     pr_response_add_err(R_501, "Login incorrect.");
     return ERROR(cmd);
@@ -1767,7 +1805,7 @@ MODRET auth_user(cmd_rec *cmd) {
       remove_config(cmd->server->conf, C_USER, FALSE);
       remove_config(cmd->server->conf, C_PASS, FALSE);
 
-      log_pri(PR_LOG_NOTICE, "USER %s (Login failed): Not a UserAlias.",
+      pr_log_pri(PR_LOG_NOTICE, "USER %s (Login failed): Not a UserAlias.",
         origuser);
       pr_response_send(R_530, "Login incorrect.");
 
@@ -1789,7 +1827,7 @@ MODRET auth_user(cmd_rec *cmd) {
         remove_config(cmd->server->conf, C_USER, FALSE);
         remove_config(cmd->server->conf, C_PASS, FALSE);
 
-        log_auth(PR_LOG_NOTICE, "ANON %s: Limit access denies login.",
+        pr_log_auth(PR_LOG_NOTICE, "ANON %s: Limit access denies login.",
           origuser);
         pr_response_send(R_530, "Login incorrect.");
 
@@ -1801,7 +1839,8 @@ MODRET auth_user(cmd_rec *cmd) {
       remove_config(cmd->server->conf, C_USER, FALSE);
       remove_config(cmd->server->conf, C_PASS, FALSE);
 
-      log_auth(PR_LOG_NOTICE, "USER %s: Limit access denies login.", origuser);
+      pr_log_auth(PR_LOG_NOTICE,
+        "USER %s: Limit access denies login.", origuser);
       pr_response_send(R_530, "Login incorrect.");
 
       end_login(0);
@@ -1923,8 +1962,12 @@ MODRET auth_pass(cmd_rec *cmd) {
       else
         pr_response_send(R_530, "Login incorrect.");
 
-      log_auth(PR_LOG_NOTICE, "Maximum login attempts (%u) exceeded",
+      pr_log_auth(PR_LOG_NOTICE, "Maximum login attempts (%u) exceeded",
         max_logins);
+
+      /* Generate an event about this limit being exceeded. */
+      pr_event_generate("mod_auth.max-login-attempts", session.c);
+
       end_login(0);
     }
 
@@ -2033,7 +2076,7 @@ MODRET add_anonymousgroup(cmd_rec *cmd) {
   argv = cmd->argv;
   argc = cmd->argc - 1;
 
-  acl = pr_parse_expression(cmd->tmp_pool, &argc, argv);
+  acl = pr_expr_create(cmd->tmp_pool, &argc, argv);
 
   c = add_config_param(cmd->argv[0], 0);
   c->argc = argc;
@@ -2110,7 +2153,9 @@ MODRET set_createhome(cmd_rec *cmd) {
   }
 
   /* Check the mode parameter, if present */
-  if (cmd->argc-1 >= 2) {
+  if (cmd->argc-1 >= 2 &&
+      strcmp(cmd->argv[2], "dirmode") != 0 &&
+      strcmp(cmd->argv[2], "skel") != 0) {
     char *tmp = NULL;
 
     mode = strtol(cmd->argv[2], &tmp, 8);
@@ -2210,7 +2255,7 @@ MODRET add_defaultroot(cmd_rec *cmd) {
   if (*(dir + strlen(dir) - 1) != '/')
     dir = pstrcat(cmd->tmp_pool, dir, "/", NULL);
 
-  acl = pr_parse_expression(cmd->tmp_pool, &argc, argv);
+  acl = pr_expr_create(cmd->tmp_pool, &argc, argv);
 
   c = add_config_param(cmd->argv[0], 0);
 
@@ -2252,7 +2297,7 @@ MODRET add_defaultchdir(cmd_rec *cmd) {
   if (*(dir + strlen(dir) - 1) != '/')
     dir = pstrcat(cmd->tmp_pool, dir, "/", NULL);
 
-  acl = pr_parse_expression(cmd->tmp_pool, &argc, argv);
+  acl = pr_expr_create(cmd->tmp_pool, &argc, argv);
 
   c = add_config_param(cmd->argv[0], 0);
 
@@ -2303,6 +2348,42 @@ MODRET set_loginpasswordprompt(cmd_rec *cmd) {
   return HANDLED(cmd);
 }
 
+/* usage: MaxClientsPerClass class max|"none" ["message"] */
+MODRET set_maxclientsclass(cmd_rec *cmd) {
+  int max;
+  config_rec *c;
+
+  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
+
+  if (strcasecmp(cmd->argv[2], "none") == 0)
+    max = 0;
+
+  else {
+    char *endp = NULL;
+
+    max = (int) strtol(cmd->argv[2], &endp, 10);
+
+    if ((endp && *endp) || max < 1)
+      CONF_ERROR(cmd, "max must be 'none' or a number greater than 0");
+  }
+
+  if (cmd->argc == 4) {
+    c = add_config_param(cmd->argv[0], 3, NULL, NULL, NULL);
+    c->argv[0] = pstrdup(c->pool, cmd->argv[1]);
+    c->argv[1] = pcalloc(c->pool, sizeof(unsigned int));
+    *((unsigned int *) c->argv[1]) = max;
+    c->argv[2] = pstrdup(c->pool, cmd->argv[3]);
+
+  } else {
+    c = add_config_param(cmd->argv[0], 2, NULL, NULL);
+    c->argv[0] = pstrdup(c->pool, cmd->argv[1]);
+    c->argv[1] = pcalloc(c->pool, sizeof(unsigned int));
+    *((unsigned int *) c->argv[1]) = max;
+  }
+
+  return HANDLED(cmd);
+}
+
 /* usage: MaxClients max|"none" ["message"] */
 MODRET set_maxclients(cmd_rec *cmd) {
   int max;
@@ -2620,7 +2701,7 @@ MODRET set_timeoutsession(cmd_rec *cmd) {
     int argc = cmd->argc - 3;
     char **argv = cmd->argv + 2;
 
-    acl = pr_parse_expression(cmd->tmp_pool, &argc, argv);
+    acl = pr_expr_create(cmd->tmp_pool, &argc, argv);
 
     c = add_config_param(cmd->argv[0], 0);
     c->argc = argc + 2;
@@ -2749,6 +2830,7 @@ static conftable auth_conftab[] = {
   { "GroupPassword",		set_grouppassword,		NULL },
   { "LoginPasswordPrompt",	set_loginpasswordprompt,	NULL },
   { "MaxClients",		set_maxclients,			NULL },
+  { "MaxClientsPerClass",	set_maxclientsclass,		NULL },
   { "MaxClientsPerHost",	set_maxhostclients,		NULL },
   { "MaxClientsPerUser",	set_maxuserclients,		NULL },
   { "MaxHostsPerUser",		set_maxhostsperuser,		NULL },
