@@ -24,9 +24,8 @@
  * the source code for OpenSSL in the source distribution.
  */
 
-/*
- * Unix authentication module for ProFTPD
- * $Id: mod_auth_unix.c,v 1.4 2003/02/18 17:55:22 castaglia Exp $
+/* Unix authentication module for ProFTPD
+ * $Id: mod_auth_unix.c,v 1.17 2004/07/21 23:58:10 castaglia Exp $
  */
 
 #include "conf.h"
@@ -34,20 +33,20 @@
 /* AIX has some rather stupid function prototype inconsistencies between
  * their crypt.h and stdlib.h's setkey() declarations.  *sigh*
  */
-#if defined(HAVE_CRYPT_H) && !defined(AIX4)
-#include <crypt.h>
+#if defined(HAVE_CRYPT_H) && !defined(AIX4) && !defined(AIX5)
+# include <crypt.h>
 #endif
 
 #ifdef USE_SHADOW
-#include <shadow.h>
+# include <shadow.h>
 #endif
 
 #ifdef HAVE_SYS_SECURITY_H
-#include <sys/security.h>
+# include <sys/security.h>
 #endif
 
 #ifdef HAVE_KRB_H
-#include <krb.h>
+# include <krb.h>
 #endif
 
 #if defined(HAVE_HPSECURITY_H) || defined(HPUX10) || defined(HPUX11)
@@ -58,7 +57,7 @@
 #endif /* HAVE_HPSECURITY_H or HPUX10 or HPUX11 */
 
 #if defined(HAVE_PROT_H) || defined(COMSEC)
-#include <prot.h>
+# include <prot.h>
 #endif
 
 #ifdef USE_SIA
@@ -70,6 +69,16 @@
 # endif
 #endif /* USE_SIA */
 
+#ifdef CYGWIN
+typedef void *HANDLE;
+typedef unsigned long DWORD;
+# define INVALID_HANDLE_VALUE (HANDLE)(-1)
+# define WINAPI __stdcall
+DWORD WINAPI GetVersion(void);
+extern HANDLE cygwin_logon_user (const struct passwd *, const char *);
+extern void cygwin_set_impersonation_token (const HANDLE);
+#endif /* CYGWIN */
+
 #ifdef SETGRENT_VOID
 # define RETSETGRENTTYPE	void
 #else
@@ -137,7 +146,7 @@ static void p_setpwent(void) {
 
   else
     if ((pwdf = fopen(PASSWD,"r")) == NULL)
-      log_pri(PR_LOG_ERR, "Unable to open password file %s for reading: %s",
+      pr_log_pri(PR_LOG_ERR, "Unable to open password file %s for reading: %s",
         PASSWD, strerror(errno));
 }
 
@@ -154,7 +163,7 @@ static RETSETGRENTTYPE p_setgrent(void) {
 
   else
     if ((grpf = fopen(GROUP,"r")) == NULL)
-      log_pri(PR_LOG_ERR, "Unable to open group file %s for reading: %s",
+      pr_log_pri(PR_LOG_ERR, "Unable to open group file %s for reading: %s",
         GROUP, strerror(errno));
 
 #ifndef SETGRENT_VOID
@@ -237,8 +246,7 @@ static struct group *p_getgrgid(gid_t gid) {
   return gr;
 }
 
-inline static int _compare_uid(idmap_t *m1, idmap_t *m2)
-{
+inline static int _compare_uid(idmap_t *m1, idmap_t *m2) {
   if (m1->id.uid < m2->id.uid)
     return -1;
 
@@ -248,8 +256,7 @@ inline static int _compare_uid(idmap_t *m1, idmap_t *m2)
   return 0;
 }
 
-inline static int _compare_gid(idmap_t *m1, idmap_t *m2)
-{
+inline static int _compare_gid(idmap_t *m1, idmap_t *m2) {
   if (m1->id.gid < m2->id.gid)
     return -1;
 
@@ -259,23 +266,21 @@ inline static int _compare_gid(idmap_t *m1, idmap_t *m2)
   return 0;
 }
 
-inline static int _compare_id(xaset_t **table, idauth_t id, idauth_t idcomp)
-{
+inline static int _compare_id(xaset_t **table, idauth_t id, idauth_t idcomp) {
   if (table == uid_table)
     return id.uid == idcomp.uid;
   else
     return id.gid == idcomp.gid;
 }
 
-static idmap_t *_auth_lookup_id(xaset_t **id_table, idauth_t id)
-{
+static idmap_t *_auth_lookup_id(xaset_t **id_table, idauth_t id) {
   int hash = ((id_table == uid_table) ? id.uid : id.gid) % HASH_TABLE_SIZE;
   idmap_t *m;
 
   if (!id_table[hash])
-    id_table[hash] = xaset_create(permanent_pool, (id_table == uid_table) ?
-                                  (XASET_COMPARE)_compare_uid :
-                                  (XASET_COMPARE)_compare_gid);
+    id_table[hash] = xaset_create(session.pool,
+      (id_table == uid_table) ? (XASET_COMPARE)_compare_uid :
+      (XASET_COMPARE)_compare_gid);
 
   for (m = (idmap_t *) id_table[hash]->xas_list; m; m = m->next) {
     if (_compare_id(id_table, m->id, id))
@@ -284,7 +289,7 @@ static idmap_t *_auth_lookup_id(xaset_t **id_table, idauth_t id)
 
   if (!m || !_compare_id(id_table, m->id, id)) {
     /* Isn't in the table */
-    m = (idmap_t *) pcalloc(id_table[hash]->mempool, sizeof(idmap_t));
+    m = (idmap_t *) pcalloc(id_table[hash]->pool, sizeof(idmap_t));
 
     if (id_table == uid_table)
       m->id.uid = id.uid;
@@ -297,24 +302,22 @@ static idmap_t *_auth_lookup_id(xaset_t **id_table, idauth_t id)
   return m;
 }
 
-MODRET pw_setpwent(cmd_rec *cmd)
-{
+MODRET pw_setpwent(cmd_rec *cmd) {
   if (PERSISTENT_PASSWD)
     p_setpwent();
   else
     setpwent();
 
-  return HANDLED(cmd);
+  return DECLINED(cmd);
 }
 
-MODRET pw_endpwent(cmd_rec *cmd)
-{
+MODRET pw_endpwent(cmd_rec *cmd) {
   if (PERSISTENT_PASSWD)
     p_endpwent();
   else
     endpwent();
 
-  return HANDLED(cmd);
+  return DECLINED(cmd);
 }
 
 MODRET pw_setgrent(cmd_rec *cmd) {
@@ -323,21 +326,19 @@ MODRET pw_setgrent(cmd_rec *cmd) {
   else
     setgrent();
 
-  return HANDLED(cmd);
+  return DECLINED(cmd);
 }
 
-MODRET pw_endgrent(cmd_rec *cmd)
-{
+MODRET pw_endgrent(cmd_rec *cmd) {
   if (PERSISTENT_GROUP)
     p_endgrent();
   else
     endgrent();
 
-  return HANDLED(cmd);
+  return DECLINED(cmd);
 }
 
-MODRET pw_getgrent(cmd_rec *cmd)
-{
+MODRET pw_getgrent(cmd_rec *cmd) {
   struct group *gr;
 
   if (PERSISTENT_GROUP)
@@ -345,14 +346,10 @@ MODRET pw_getgrent(cmd_rec *cmd)
   else
     gr = getgrent();
 
-  if (gr)
-    return mod_create_data(cmd,gr);
-  else
-    return ERROR(cmd);
+  return gr ? mod_create_data(cmd, gr) : DECLINED(cmd);
 }
 
-MODRET pw_getpwent(cmd_rec *cmd)
-{
+MODRET pw_getpwent(cmd_rec *cmd) {
   struct passwd *pw;
 
   if (PERSISTENT_PASSWD)
@@ -360,14 +357,10 @@ MODRET pw_getpwent(cmd_rec *cmd)
   else
     pw = getpwent();
 
-  if (pw)
-    return mod_create_data(cmd,pw);
-  else
-    return ERROR(cmd);
+  return pw ? mod_create_data(cmd, pw) : DECLINED(cmd);
 }
 
-MODRET pw_getpwuid(cmd_rec *cmd)
-{
+MODRET pw_getpwuid(cmd_rec *cmd) {
   struct passwd *pw;
   uid_t uid;
 
@@ -377,14 +370,10 @@ MODRET pw_getpwuid(cmd_rec *cmd)
   else
     pw = getpwuid(uid);
 
-  if (pw)
-    return mod_create_data(cmd,pw);
-  else
-    return ERROR(cmd);
+  return pw ? mod_create_data(cmd, pw) : DECLINED(cmd);
 }
 
-MODRET pw_getpwnam(cmd_rec *cmd)
-{
+MODRET pw_getpwnam(cmd_rec *cmd) {
   struct passwd *pw;
   const char *name;
 
@@ -394,14 +383,10 @@ MODRET pw_getpwnam(cmd_rec *cmd)
   else
     pw = getpwnam(name);
 
-  if (pw)
-    return mod_create_data(cmd,pw);
-  else
-    return ERROR(cmd);
+  return pw ? mod_create_data(cmd, pw) : DECLINED(cmd);
 }
 
-MODRET pw_getgrnam(cmd_rec *cmd)
-{
+MODRET pw_getgrnam(cmd_rec *cmd) {
   struct group *gr;
   const char *name;
 
@@ -411,14 +396,10 @@ MODRET pw_getgrnam(cmd_rec *cmd)
   else
     gr = getgrnam(name);
 
-  if (gr)
-    return mod_create_data(cmd,gr);
-  else
-    return ERROR(cmd);
+  return gr ? mod_create_data(cmd, gr) : DECLINED(cmd);
 }
 
-MODRET pw_getgrgid(cmd_rec *cmd)
-{
+MODRET pw_getgrgid(cmd_rec *cmd) {
   struct group *gr;
   gid_t gid;
 
@@ -428,17 +409,12 @@ MODRET pw_getgrgid(cmd_rec *cmd)
   else
     gr = getgrgid(gid);
 
-  if (gr)
-    return mod_create_data(cmd,gr);
-  else
-    return ERROR(cmd);
+  return gr ? mod_create_data(cmd, gr) : DECLINED(cmd);
 }
 
 #ifdef USE_SHADOW
-static char *_get_pw_info(pool *p, const char *u,
-                          time_t *lstchg, time_t *min, time_t *max,
-                          time_t *warn, time_t *inact, time_t *expire)
-{
+static char *_get_pw_info(pool *p, const char *u, time_t *lstchg, time_t *min,
+    time_t *max, time_t *warn, time_t *inact, time_t *expire) {
   struct spwd *sp;
   char *cpw = NULL;
 
@@ -472,7 +448,7 @@ static char *_get_pw_info(pool *p, const char *u,
       *expire = SP_CVT_DAYS(sp->sp_expire);
 #endif /* HAVE_SPWD_SP_EXPIRE */
   }
-#ifdef AUTO_SHADOW
+#ifdef USE_AUTO_SHADOW
   else {
     struct passwd *pw;
 
@@ -480,28 +456,38 @@ static char *_get_pw_info(pool *p, const char *u,
     PRIVS_RELINQUISH
 
     if ((pw = getpwnam(u)) != NULL) {
-      cpw = pstrdup(p,pw->pw_passwd);
-      if (lstchg) *lstchg = (time_t)-1;
-      if (min) *min = (time_t)-1;
-      if (max) *max = (time_t)-1;
-      if (warn) *warn = (time_t)-1;
-      if (inact) *inact = (time_t)-1;
-      if (expire) *expire = (time_t)-1;
+      cpw = pstrdup(p, pw->pw_passwd);
+
+      if (lstchg)
+        *lstchg = (time_t) -1;
+
+      if (min)
+        *min = (time_t) -1;
+
+      if (max)
+        *max = (time_t) -1;
+
+      if (warn)
+        *warn = (time_t) -1;
+
+      if (inact)
+        *inact = (time_t) -1;
+
+      if (expire)
+        *expire = (time_t) -1;
     }
   }
 #else
   endspent();
   PRIVS_RELINQUISH
-#endif /* AUTO_SHADOW */
+#endif /* USE_AUTO_SHADOW */
   return cpw;
 }
 
 #else /* USE_SHADOW */
 
-static char *_get_pw_info(pool *p, const char *u,
-                          time_t *lstchg, time_t *min, time_t *max,
-                          time_t *warn, time_t *inact, time_t *expire)
-{
+static char *_get_pw_info(pool *p, const char *u, time_t *lstchg, time_t *min,
+    time_t *max, time_t *warn, time_t *inact, time_t *expire) {
   char *cpw = NULL;
 #if defined(HAVE_GETPRPWENT) || defined(COMSEC)
   struct pr_passwd *prpw;
@@ -510,7 +496,7 @@ static char *_get_pw_info(pool *p, const char *u,
   struct passwd *pw;
 #endif
 
- /* some platforms (i.e. bsd) provide "transparent" shadowing, which
+ /* Some platforms (i.e. BSD) provide "transparent" shadowing, which
   * requires that we are root in order to have the password member
   * filled in.
   */
@@ -536,12 +522,24 @@ static char *_get_pw_info(pool *p, const char *u,
 
   if (pw) {
     cpw = pstrdup(p, pw->pw_passwd);
-    if (lstchg) *lstchg = (time_t)-1;
-    if (min) *min = (time_t)-1;
-    if (max) *max = (time_t)-1;
-    if (warn) *warn = (time_t)-1;
-    if (inact) *inact = (time_t)-1;
-    if (expire) *expire = (time_t)-1;
+
+    if (lstchg)
+      *lstchg = (time_t) -1;
+
+    if (min)
+      *min = (time_t) -1;
+
+    if (max)
+      *max = (time_t) -1;
+
+    if (warn)
+      *warn = (time_t) -1;
+
+    if (inact)
+      *inact = (time_t) -1;
+
+    if (expire)
+      *expire = (time_t) -1;
   }
 
   endpwent();
@@ -554,16 +552,28 @@ static char *_get_pw_info(pool *p, const char *u,
   endprpwent();
   setprpwent();
 
-  prpw = getprpwnam((char *)u);
+  prpw = getprpwnam((char *) u);
 
   if (prpw) {
     cpw = pstrdup(p, prpw->ufld.fd_encrypt);
-    if (lstchg) *lstchg = (time_t)-1;
-    if (min) *min = prpw->ufld.fd_min;
-    if (max) *max = (time_t)-1;
-    if (warn) *warn = (time_t)-1;
-    if (inact) *inact = (time_t)-1;
-    if (expire) *expire = prpw->ufld.fd_expire;
+
+    if (lstchg)
+      *lstchg = (time_t) -1;
+
+    if (min)
+      *min = prpw->ufld.fd_min;
+
+    if (max)
+      *max = (time_t) -1;
+
+    if (warn)
+      *warn = (time_t) -1;
+
+    if (inact)
+      *inact = (time_t) -1;
+
+    if (expire)
+      *expire = prpw->ufld.fd_expire;
   }
 
   endprpwent();
@@ -587,12 +597,12 @@ static char *_get_ppw_info(pool *p, const char *u) {
 
   pw = p_getpwnam(u);
   if (pw)
-    cpw = pstrdup(p,pw->pw_passwd);
+    cpw = pstrdup(p, pw->pw_passwd);
 
   return cpw;
 }
 
-/* high-level auth handlers
+/* High-level auth handlers
  */
 
 /* cmd->argv[0] : user name
@@ -602,36 +612,38 @@ static char *_get_ppw_info(pool *p, const char *u) {
 MODRET pw_auth(cmd_rec *cmd) {
   time_t now;
   char *cpw;
-  time_t lstchg = -1,max = -1,inact = -1,disable = -1;
+  time_t lstchg = -1, max = -1, inact = -1, disable = -1;
   const char *name;
 
   name = cmd->argv[0];
   time(&now);
 
   if (persistent_passwdf)
-    cpw = _get_ppw_info(cmd->tmp_pool,name);
+    cpw = _get_ppw_info(cmd->tmp_pool, name);
   else
-    cpw = _get_pw_info(cmd->tmp_pool,name,&lstchg,NULL,&max,NULL,&inact,&disable);
+    cpw = _get_pw_info(cmd->tmp_pool, name, &lstchg, NULL, &max, NULL, &inact,
+      &disable);
 
   if (!cpw)
-    return ERROR_INT(cmd, PR_AUTH_NOPWD);
+    return DECLINED(cmd);
 
-  if (auth_check(cmd->tmp_pool,cpw,cmd->argv[0],cmd->argv[1]))
+  if (auth_check(cmd->tmp_pool, cpw, cmd->argv[0], cmd->argv[1]))
     return ERROR_INT(cmd, PR_AUTH_BADPWD);
 
-  if (lstchg > (time_t)0 && max > (time_t)0 &&
-     inact > (time_t)0)
+  if (lstchg > (time_t) 0 &&
+      max > (time_t) 0 &&
+      inact > (time_t)0)
     if (now > lstchg + max + inact)
       return ERROR_INT(cmd, PR_AUTH_AGEPWD);
 
-  if (disable > (time_t)0 && now > disable)
+  if (disable > (time_t) 0 &&
+      now > disable)
     return ERROR_INT(cmd, PR_AUTH_DISABLEDPWD);
 
   return HANDLED(cmd);
 }
 
-/*
- * cmd->argv[0] = hashed password,
+/* cmd->argv[0] = hashed password,
  * cmd->argv[1] = user,
  * cmd->argv[2] = cleartext
  */
@@ -650,7 +662,7 @@ MODRET pw_check(cmd_rec *cmd) {
 
 #ifdef COMSEC
   if (iscomsec()) {
-    if (strcmp(bigcrypt(pw, cpw), cpw) != 0)
+    if (strcmp(bigcrypt((char *) pw, (char *) cpw), cpw) != 0)
       return ERROR(cmd);
 
   } else {
@@ -667,17 +679,17 @@ MODRET pw_check(cmd_rec *cmd) {
   PRIVS_ROOT
   if ((res = sia_ses_init(&ent, 1, info, NULL, user, NULL, 0,
       NULL)) != SIASUCCESS) {
-    log_auth(PR_LOG_NOTICE, "sia_ses_init() returned %d for user '%s'", res,
+    pr_log_auth(PR_LOG_NOTICE, "sia_ses_init() returned %d for user '%s'", res,
       user);
 
   } else {
 
     if ((res = sia_ses_authent(NULL, pw, ent)) != SIASUCCESS)
-      log_auth(PR_LOG_NOTICE, "sia_ses_authent() returned %d for user '%s'",
+      pr_log_auth(PR_LOG_NOTICE, "sia_ses_authent() returned %d for user '%s'",
         res, user);
 
     if ((res = sia_ses_release(&ent)) != SIASUCCESS)
-      log_auth(PR_LOG_NOTICE, "sia_ses_release() returned %d", res);
+      pr_log_auth(PR_LOG_NOTICE, "sia_ses_release() returned %d", res);
   }
   PRIVS_RELINQUISH
 
@@ -685,6 +697,51 @@ MODRET pw_check(cmd_rec *cmd) {
     return ERROR(cmd);
 
 #else /* !USE_SIA */
+
+# ifdef CYGWIN
+  /* We have to do special Windows NT voodoo with Cygwin in order to be
+   * able to switch UID/GID. More info at
+   * http://cygwin.com/cygwin-ug-net/ntsec.html#NTSEC-SETUID
+   */
+  if (GetVersion() < 0x80000000) {
+    cmd_rec *tmp_cmd = NULL;
+    modret_t *mr = NULL;
+    struct passwd *pwent = NULL;
+    HANDLE token;
+
+    /* A struct passwd * is needed.  To look one up via pw_getpwnam(), though,
+     * we'll need a cmd_rec.
+     */
+    tmp_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, cmd->argv[1]);
+
+    /* pw_getpwnam() returns a MODRET, so we need to handle that.  Yes, this
+     * might have been easier if we'd used auth_getpwnam(), but that would
+     * dispatch through other auth modules, which is _not_ what we want.
+     */
+    mr = pw_getpwnam(tmp_cmd);
+
+    /* Note: we don't handle the case where pw_getpwnam() returns anything
+     * other than HANDLED at the moment.
+     */
+
+    if (MODRET_ISHANDLED(mr) && MODRET_HASDATA(mr)) {
+      pwent = mr->data;
+
+      if ((token = cygwin_logon_user((const struct passwd *) pwent,
+          pw)) == INVALID_HANDLE_VALUE) {
+        pr_log_pri(PR_LOG_NOTICE, "error authenticating Cygwin user: %s",
+          strerror(errno));
+        return ERROR(cmd);
+      }
+
+      cygwin_set_impersonation_token(token);
+
+    } else
+      return DECLINED(cmd);
+
+  } else
+# endif /* CYGWIN */
+
   if (strcmp(crypt(pw, cpw), cpw) != 0)
     return ERROR(cmd);
 #endif /* USE_SIA */
@@ -696,7 +753,7 @@ MODRET pw_check(cmd_rec *cmd) {
   return HANDLED(cmd);
 }
 
-MODRET pw_uid_name(cmd_rec *cmd) {
+MODRET pw_uid2name(cmd_rec *cmd) {
   idmap_t *m;
   idauth_t id;
   struct passwd *pw;
@@ -705,7 +762,7 @@ MODRET pw_uid_name(cmd_rec *cmd) {
   m = _auth_lookup_id(uid_table, id);
 
   if (!m->name) {
-    /* wasn't cached, so perform a lookup */
+    /* Wasn't cached, so perform a lookup */
 
     if (PERSISTENT_PASSWD)
       pw = p_getpwuid(id.uid);
@@ -713,20 +770,18 @@ MODRET pw_uid_name(cmd_rec *cmd) {
       pw = getpwuid(id.uid);
 
     if (pw) {
-      m->name = pstrdup(permanent_pool, pw->pw_name);
-    } else {
-      char buf[10] = {'\0'};
-
-      snprintf(buf, sizeof(buf), "%lu", (ULONG)id.uid);
-      m->name = pstrdup(permanent_pool, buf);
+      m->name = pstrdup(session.pool ? session.pool : permanent_pool,
+        pw->pw_name);
+      return mod_create_data(cmd, m->name);
     }
+
+    return DECLINED(cmd);
   }
 
   return mod_create_data(cmd, m->name);
 }
 
-MODRET pw_gid_name(cmd_rec *cmd)
-{
+MODRET pw_gid2name(cmd_rec *cmd) {
   idmap_t *m;
   idauth_t id;
   struct group *gr;
@@ -741,26 +796,19 @@ MODRET pw_gid_name(cmd_rec *cmd)
     else
       gr = getgrgid(id.gid);
 
-    if (gr)
-      m->name = pstrdup(permanent_pool, gr->gr_name);
-    else {
-      char buf[10] = {'\0'};
-
-      /* removed cast to unsigned long long here, as it presents a problem
-       * passed to snprintf because there is no ansi standard for the format
-       * string modifier used for long long (is it %llu or %Lu, etc?)
-       * jss 2/21/01
-       */
-      snprintf(buf, sizeof(buf), "%lu", (ULONG)id.gid);
-      m->name = pstrdup(permanent_pool, buf);
+    if (gr) {
+      m->name = pstrdup(session.pool ? session.pool : permanent_pool,
+        gr->gr_name);
+      return mod_create_data(cmd, m->name);
     }
+
+    return DECLINED(cmd);
   }
 
   return mod_create_data(cmd, m->name);
 }
 
-MODRET pw_name_uid(cmd_rec *cmd)
-{
+MODRET pw_name2uid(cmd_rec *cmd) {
   struct passwd *pw;
   const char *name;
 
@@ -771,15 +819,11 @@ MODRET pw_name_uid(cmd_rec *cmd)
   else
     pw = getpwnam(name);
 
-  if (pw)
-    return mod_create_data(cmd,(void*)pw->pw_uid);
-  return ERROR(cmd);
+  return pw ? mod_create_data(cmd, (void *) pw->pw_uid) : DECLINED(cmd);
 }
 
-MODRET pw_name_gid(cmd_rec *cmd)
-{
+MODRET pw_name2gid(cmd_rec *cmd) {
   struct group *gr;
-
   const char *name;
 
   name = cmd->argv[0];
@@ -789,9 +833,7 @@ MODRET pw_name_gid(cmd_rec *cmd)
   else
     gr = getgrnam(name);
 
-  if (gr)
-    return mod_create_data(cmd,(void*)gr->gr_gid);
-  return ERROR(cmd);
+  return gr ? mod_create_data(cmd, (void *) gr->gr_gid) : DECLINED(cmd);
 }
 
 /* cmd->argv[0] = name
@@ -827,36 +869,35 @@ MODRET pw_getgroups(cmd_rec *cmd) {
 
   name = (char *) cmd->argv[0];
 
-  /* check for NULL values */
+  /* Check for NULL values. */
   if (cmd->argv[1])
     gids = (array_header *) cmd->argv[1];
 
   if (cmd->argv[2])
     groups = (array_header *) cmd->argv[2];
 
-  /* retrieve the necessary info */
+  /* Retrieve the necessary info. */
   if (!name || !(pw = my_getpwnam(name)))
-    return mod_create_error(cmd, -1);
+    return DECLINED(cmd);
 
-  /* populate the first group ID and name
-   */
+  /* Populate the first group ID and name. */
   if (gids)
     *((gid_t *) push_array(gids)) = pw->pw_gid;
 
   if (groups && (gr = my_getgrgid(pw->pw_gid)) != NULL)
-    *((char **) push_array(groups)) = pstrdup(permanent_pool, gr->gr_name);
+    *((char **) push_array(groups)) = pstrdup(session.pool, gr->gr_name);
 
   my_setgrent();
 
-  /* this is where things get slow, expensive, and ugly.  Loop through
+  /* This is where things get slow, expensive, and ugly.  Loop through
    * everything, checking to make sure we haven't already added it.
    */
   while ((gr = my_getgrent()) != NULL && gr->gr_mem) {
 
-    /* loop through each member name listed */
+    /* Loop through each member name listed */
     for (gr_member = gr->gr_mem; *gr_member; gr_member++) {
 
-      /* if it matches the given username... */
+      /* If it matches the given username... */
       if (!strcmp(*gr_member, pw->pw_name)) {
 
         /* ...add the GID and name */
@@ -864,7 +905,7 @@ MODRET pw_getgroups(cmd_rec *cmd) {
           *((gid_t *) push_array(gids)) = gr->gr_gid;
 
         if (groups && pw->pw_gid != gr->gr_gid)
-          *((char **) push_array(groups)) = pstrdup(permanent_pool,
+          *((char **) push_array(groups)) = pstrdup(session.pool,
             gr->gr_name);
       }
     }
@@ -896,6 +937,11 @@ MODRET set_persistentpasswd(cmd_rec *cmd) {
 /* Initialization routines
  */
 
+static void auth_unix_sess_exit(void) {
+  auth_endpwent(session.pool);
+  auth_endgrent(session.pool);
+}
+
 static int auth_unix_init(void) {
   memset(uid_table, 0 ,sizeof(uid_table));
   memset(gid_table, 0, sizeof(gid_table));
@@ -926,6 +972,7 @@ static int auth_unix_sess_init(void) {
     p_setgrent();
   }
 
+  pr_exit_register_handler(auth_unix_sess_exit);
   return 0;
 }
 
@@ -950,10 +997,10 @@ static authtable auth_unix_authtab[] = {
   { 0,  "getgrgid",     pw_getgrgid },
   { 0,  "auth",         pw_auth	},
   { 0,  "check",	pw_check },
-  { 0,  "uid_name",	pw_uid_name },
-  { 0,  "gid_name",	pw_gid_name },
-  { 0,  "name_uid",	pw_name_uid },
-  { 0,  "name_gid",	pw_name_gid },
+  { 0,  "uid_name",	pw_uid2name },
+  { 0,  "gid_name",	pw_gid2name },
+  { 0,  "name_uid",	pw_name2uid },
+  { 0,  "name_gid",	pw_name2gid },
   { 0,  "getgroups",	pw_getgroups },
   { 0,  NULL }
 };
